from functools import lru_cache
from math import factorial


@lru_cache
def tetrate(a, b):
    if not (b <= 2 or a * b < 10):
        raise MemoryError('Value too large to compute')
    if b == 0:
        return 1
    return a ** tetrate(a, b - 1)


@lru_cache
def double_factorial(a: int):
    if a < 0:
        raise ValueError('Argument to double factorial must be a positive integer')
    if a == 0 or a == 1:
        return 1
    return a * double_factorial(a - 2)


@lru_cache
def superfactorial(a: int):
    if a < 0:
        raise ValueError('Argument to superfactorial must be a positive integer')
    if a == 0:
        return 1
    return factorial(a) * superfactorial(a - 1)

@lru_cache
def hyperfactorial(a: int):
    if a < 0:
        raise ValueError('Argument to hyperfactorial must be a positive integer')
    if a == 0:
        return 1
    return (a ** a) * hyperfactorial(a - 1)

### FUTURE FEATURES ###

# def series(start: int, end: int, function):
#     pass



from typing import Callable, List, Literal, Union

from .utils.priority import Priority
from .exceptions import ReduceTreeError
from .node import Node
from math import ceil


class Operator(Node):
    def __init__(self,
                 symbol: str,
                 func: Callable,
                 priority: Union[int, Priority],
                 kind: Union[Literal['pre'], Literal['in'], Literal['post']],
                 operands: int,
                 **kwargs):
        super().__init__(**kwargs)
        self.symbol: str = symbol
        self.priority: int = priority.value if isinstance(priority, Priority) else priority
        self.kind = kind
        self.operands = operands
        self.func = func
        self.width = len(symbol)
        self.inner_width = len(symbol)

    def __call__(self) -> Union[int, float]:
        if self.children is None:
            raise ReduceTreeError('Tree is not complete')
        return self.func(*map(lambda op: op.value, self.children))

    def __str__(self) -> str:
        return self.symbol

    def display(self):
        if len(self.symbol) <= (self.width - abs(self.bal_coef)):
            if self.bal_coef > 0:
                return ' ' * self.left_pad + f'{self.symbol:^{self.width - abs(self.bal_coef)}}' + ' ' * (abs(self.bal_coef) + self.right_pad)
            else:
                return ' ' * (self.left_pad + abs(self.bal_coef)) + f'{self.symbol:^{self.width - abs(self.bal_coef)}}' + ' ' * self.right_pad
        elif self.bal_coef > 0:
            return ' ' * self.left_pad + f'{self.symbol:<{self.width}}' + ' ' * self.right_pad
        elif self.bal_coef < 0:
            return ' ' * self.left_pad + f'{self.symbol:>{self.width}}' + ' ' * self.right_pad
        return ' ' * self.left_pad + f'{self.symbol:^{self.width}}' + ' ' * self.right_pad

    def __repr__(self) -> str:
        return f'Operator({self.symbol}, kind=\'{self.kind}\', operands={self.operands})'

    def isFull(self) -> bool:
        return len(self.children) == self.operands

    # def copy(self):
    #     return Operator(symbol=self.symbol, fun)



# from typing import Dict, List, cast

# from .utils.io_utils import clear_console

# from .node import Node
# from .operand_ import Operand
# from .operator_ import Operator

# from .utils.string_utils import concatenate_horizontally, simplify_expression
# from .parser import Parser
# from .lexer import Lexer
# from .tokenizer import Tokenizer
# import os
# from copy import deepcopy
# from time import sleep


# class Scene:
#     def __init__(self,
#     parse_tree_class,
#                  token_lookup: Dict,
#                  implicit_operator: Operator,
#                  seconds_per_frame: int = 1) -> None:
#         self.parse_tree_class = parse_tree_class
#         self.token_lookup = token_lookup
#         self.implicit_operator = implicit_operator
#         self.seconds_per_frame = seconds_per_frame

#     def stencil(self, expression, tokens=None, lexed=None, parsed=None, tree=None) -> None:
#         clear_console()
#         if tokens is None:
#             process = 'TOKENIZATION'
#         elif lexed is None:
#             process = 'LEXING'
#         elif parsed is None:
#             process = 'PARSING'
#         elif tree is None:
#             process = 'BUILDING PARSE TREE'
#         else:
#             process = 'REDUCING PARSE TREE'
#         print(f'CURRENT PROCESS: {process}')
#         print()
#         print(f'Expression: {expression}')
#         if tokens is not None:
#             print()
#             print('TOKENS >>>')
#             print(tokens)
#         if lexed is not None:
#             print()
#             print('LEXED TOKENS >>>')
#             print(lexed)
#             print()
#         if parsed is not None:
#             print('PARSED TOKENS >>>')
#             print(parsed)
#             print()
#         if tree is not None:
#             print('COMPLETE PARSE TREE >>>')
#             print(tree)
#             print()

#     def animate_tokenization(self, expression):
#         tokens: List[str] = []
#         for i, j, t in Tokenizer(token_lookup=self.token_lookup).tokenize(expression=expression, kind='lazy'):
#             self.stencil(expression)
#             print(' ' * (i + 12) + '\u2191' + ' ' * (j - i - 1) + '\u2191')
#             print(' ' * (i + 12) + 'i' + ' ' * (j - i - 1) + 'j')
#             print(t)
#             tokens = t
#             sleep(self.seconds_per_frame)
#         return tokens

#     def animate_lexing(self, expression, tokens):
#         lexed = []
#         prefix = 1
#         for o, c, l, t in Lexer(token_lookup=self.token_lookup, implicit_operator=self.implicit_operator).lex(tokens=tokens, kind='lazy'):
#             self.stencil(expression, tokens)
#             arrow = '\u2191'
#             print(' ' * prefix + f'{arrow:^{len(t) + 2}}')
#             print(f'(: {o}; ): {c}')
#             print(l)
#             lexed = l
#             sleep(self.seconds_per_frame)
#             prefix += len(t) + 4
#         return lexed

#     def animate_parsing(self, lexed):
#         return Parser(token_lookup=self.token_lookup, implicit_operator=self.implicit_operator).parse(lexed)

#     def animate_building(self, expression, tokens, lexed, parsed):
#         tree = None
#         for i in range(1, len(parsed) + 1):
#             self.stencil(expression, tokens, lexed, parsed)
#             tmp_parsed = deepcopy(parsed[:i])
#             t = self.parse_tree_class()
#             t.build_sub_tree(tmp_parsed)
#             print(t)
#             tree = t
#             sleep(self.seconds_per_frame)
#         return tree

#     def animate_reduction(self, expression, tokens, lexed, parsed, tree, tree_repr):
#         reduced_tree = tree
#         prev_display = ''
#         for t in self.dfs_reduce(node=tree.root, tree=tree):
#             if str(t) == prev_display:
#                 continue
#             self.stencil(expression, tokens, lexed, parsed, tree_repr)
#             prev_display = str(t)
#             print(prev_display)
#             sleep(self.seconds_per_frame)
#         self.stencil(expression, tokens, lexed, parsed, tree_repr)
#         reduced_tree.root = Operand(value=reduced_tree.root())
#         print(str(reduced_tree))
#         return reduced_tree.root

#     def dfs_reduce(self, node: Node, tree):
#         for c in node.children:
#             yield from self.dfs_reduce(c, tree)
#         for i, c in enumerate(node.children):
#             node.children[i] = Operand(value=c())
#             yield tree

#     def animate(self, expression):
#         expression_stripped = simplify_expression(expression)
#         tokens = self.animate_tokenization(expression_stripped)
#         lexed = self.animate_lexing(expression_stripped, tokens)
#         parsed = self.animate_parsing(lexed)
#         t = self.animate_building(expression_stripped, tokens, lexed, parsed)
#         complete_tree = str(t)
#         result = self.animate_reduction(
#             expression, tokens, lexed, parsed, t, complete_tree)
#         self.stencil(expression_stripped, tokens, lexed, parsed, complete_tree)
#         print('REDUCED PARSE TREE >>>')
#         print(result)
#         print()
#         print('EVALUATION >>>')
#         result = t.evaluate()
#         print(expression, '=', round(result, 5))
#         return result



from typing import Any, Callable, List, Literal, Optional, cast

from .exceptions import TreeReprError, UpdateTreeError
from .temp_node import Temp_Node
from .node import Node
from math import ceil


class Tree:
    def __init__(self,
                 root: Node = None,
                 padding: int = 1,
                 left_branch: str = '/',
                 middle_branch: str = '|',
                 right_branch: str = '\\'):
        self.root: Optional[Node] = root
        self.currentPointer: Optional[Node] = None
        self.left_branch: str = left_branch
        self.middle_branch: str = middle_branch
        self.right_branch: str = right_branch
        self.padding: int = padding

    def insert(self, node: Node):
        if self.root is None or self.currentPointer is None:
            self.root = self.currentPointer = node
        elif not self.currentPointer.isFull():
            node.parent = self.currentPointer
            self.currentPointer.children.append(node)
            if self.currentPointer.isFull():
                self.currentPointer = node
        else:
            if node > self.currentPointer:
                node.parent = self.currentPointer
                self.currentPointer.children[-1].parent = node
                node.children.append(self.currentPointer.children.pop())
                self.currentPointer.children.append(node)
                self.currentPointer = node
            else:
                self.float_node(node=node)

    def float_node(self, node: Node):
        if self.currentPointer is None:
            self.root.parent = node
            node.children.append(self.root)
            self.root = node
            self.currentPointer = self.root
        elif node <= self.currentPointer:
            self.currentPointer = self.currentPointer.parent
            self.float_node(node=node)
        else:
            node.parent = self.currentPointer
            self.currentPointer.children[-1].parent = node
            node.children.append(self.currentPointer.children.pop())
            self.currentPointer.children.append(node)
            self.currentPointer = node

    def reset(self):
        self.root = None
        self.currentPointer = None
        return self

    def update_values(self, node: Node):
        for c in node.children:
            self.update_values(node=c)
        node.value = node()

    def update_dims(self, node: Node):
        if len(node.children) > 0:
            for c in node.children:
                self.update_dims(c)
            node.sub_width = sum(map(lambda n: n.width, node.children)
                                 ) + (len(node.children) - 1) * self.padding
            node.width = max(node.inner_width, node.sub_width)
            node.height = max(map(lambda n: n.height, node.children)) + 1
            if node.inner_width <= node.sub_width:
                node.bal_coef = node.get_balance_coefficient()

    def update_paddings(self, node: Node):
        if node.parent is not None:
            if node.parent.inner_width > node.parent.sub_width:
                if node is node.parent.children[0]:
                    node.left_pad += ceil((node.parent.inner_width -
                                          node.parent.sub_width) / 2)
                if node is node.parent.children[-1]:
                    node.right_pad += (node.parent.inner_width -
                                       node.parent.sub_width) // 2
            if node is node.parent.children[0]:
                node.left_pad += node.parent.left_pad
            if node is node.parent.children[-1]:
                node.right_pad += node.parent.right_pad
        for c in node.children:
            self.update_paddings(c)

    def reset_paddings(self, node: Node = None):
        if self.root is not None:
            if node is None:
                node = self.root
            node.left_pad = 0
            node.right_pad = 0
            for c in node.children:
                self.reset_paddings(node=c)
        return self

    def generate_tmp_nodes(self, node: Node, depth: int = 1):
        if len(node.children) > 0:
            for c in node.children:
                self.generate_tmp_nodes(node=c, depth=depth + 1)
        elif depth < cast(Node, self.root).height:
            self.attach_tmp_node(node=node, depth=depth)

    def attach_tmp_node(self, node: Node, depth: int):
        total_height = cast(Node, self.root).height
        for _ in range(total_height - depth):
            node = Temp_Node(parent=node)

    def prune_tmp_nodes(self, node: Node):
        if len(node.children) > 0 and isinstance(node.children[0], Temp_Node):
            node.children.pop()
        for c in node.children:
            self.prune_tmp_nodes(node=c)

    def update(self, which: Literal['all', 'dimensions', 'values'] = 'all'):
        if self.root is not None:
            if which != 'dimensions':
                self.update_values(node=self.root)
            if which != 'values':
                self.update_dims(node=self.root)
                self.reset_paddings(node=self.root)
                self.update_paddings(node=self.root)
            if which not in {'all', 'dimensions', 'values'}:
                raise UpdateTreeError(
                    f'Unexpected value \'{which}\' for \'which\' parameter encountered')

    def __enter__(self):
        if self.root is None:
            raise UpdateTreeError('Root is None')
        self.generate_tmp_nodes(node=self.root)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.root is None:
            raise UpdateTreeError('Root is None')
        self.prune_tmp_nodes(node=self.root)

    def __str__(self):
        self.update(which='dimensions')
        if self.root is not None:
            current_children: List[List[Node]] = [[self.root]]
            next_children: List[List[Node]] = []
            result = ''
            with self:
                while len(current_children[0]) > 0:
                    for n_ls in current_children:
                        for n in n_ls:
                            result += n.display() + ' ' * self.padding
                            next_children.append(n.children)
                    result += '\n'
                    current_children = next_children
                    next_children = []
                    for n_ls in current_children:
                        no_of_children = len(n_ls)
                        if no_of_children == 1:
                            n = n_ls[0]
                            if isinstance(n, Temp_Node):
                                result += n.display() + ' ' * self.padding
                            else:
                                # result += f'{self.middle_branch:^{n.width}} '
                                result += n.branch(self.middle_branch) + \
                                    ' ' * self.padding
                        elif no_of_children > 1:
                            # result += f'{self.left_branch:^{n_ls[0].width}} '
                            result += n_ls[0].branch(self.left_branch) + \
                                ' ' * self.padding
                            for n in n_ls[1:-1]:
                                # result += f'{self.middle_branch:^{n.width}} '
                                result += n.branch(self.middle_branch) + \
                                    ' ' * self.padding
                            # result += f'{self.right_branch:^{n_ls[-1].width}} '
                            result += n_ls[-1].branch(self.right_branch) + \
                                ' ' * self.padding
                    result += '\n'
            # Remove 2 redundant newline characters at end of string
            return result[:-2]
        raise TreeReprError('Empty Tree Encountered')

    def bfs(self,
            callback: Callable[[Node], Any] = lambda _: None,
            current_nodes: List[Node] = None) -> None:
        if self.root is not None:
            if current_nodes is None:
                current_nodes = [self.root]
            next_nodes = []
            for c in current_nodes:
                callback(c)
                next_nodes.extend(c.children)
            if len(next_nodes) > 0:
                self.bfs(callback=callback, current_nodes=next_nodes)

    def dfs(self,
            pre_callback: Callable[[Node], Any] = lambda _: None,
            post_callback: Callable[[Node], Any] = lambda _: None,
            current_node: Node = None) -> None:
        if self.root is not None:
            if current_node is None:
                current_node = self.root
            pre_callback(current_node)
            for c in current_node.children:
                self.dfs(pre_callback=pre_callback,
                         post_callback=post_callback,
                         current_node=c)
            post_callback(current_node)

    def __call__(self):
        return self



from math import factorial, sqrt
from math import asinh, acosh, atanh
from math import sinh, cosh, tanh
from math import asin, acos, atan
from math import sin, cos, tan
from math import log10, log
from math import pi, e, tau
from typing import List, Union

from .utils.custom_functions import double_factorial, hyperfactorial, superfactorial, tetrate

from .utils.priority import Priority
from .operand_ import Operand
from .operator_ import Operator


Bundle = List[Union[Operand, Operator]]


# Basics Bundle
BASICS: Bundle = [
    Operator(symbol='+', func=lambda a, b: a + b, priority=Priority.IN_LOW, kind='in', operands=2),
    Operator(symbol='-', func=lambda a, b: a - b, priority=Priority.IN_LOW, kind='in', operands=2),
    Operator(symbol='*', func=lambda a, b: a * b, priority=Priority.IN_MID, kind='in', operands=2),
    Operator(symbol='/', func=lambda a, b: a / b, priority=Priority.IN_MID, kind='in', operands=2),
    Operator(symbol='-', func=lambda a: -a, priority=Priority.PRE, kind='pre', operands=1)
]

# Increment Bundle
INCREMENT: Bundle = [
    Operator(symbol='++', func=lambda a: a + 1, priority=Priority.POST, kind='post', operands=1),
    Operator(symbol='--', func=lambda a: a - 1, priority=Priority.POST, kind='post', operands=1)
]


# Mod Bundle
MOD: Bundle = [
    # Modulo Operator
    Operator(symbol='%', func=lambda a, b: a % b, priority=Priority.IN_MID, kind='in', operands=2),
    Operator(symbol='mod', func=lambda a, b: a % b, priority=Priority.IN_MID, kind='in', operands=2),
    Operator(symbol='modulo', func=lambda a, b: a % b, priority=Priority.IN_MID, kind='in', operands=2),

    # Modulus Operator
    Operator(symbol='abs', func=lambda a: a if a >= 0 else -a, priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='modulus', func=lambda a: a if a >= 0 else -a, priority=Priority.PRE, kind='pre', operands=1),

    # Floor Division Operator
    Operator(symbol='//', func=lambda a, b: a // b, priority=Priority.IN_MID, kind='in', operands=2)
]

# Constants Bundle
CONSTANTS: Bundle = [
    Operand(value=pi, symbol='pi'),
    Operand(value=e, symbol='e'),
    Operand(value=tau, symbol='tau'),
    Operand(value=(1 + (5 ** 0.5)) / 2, symbol='phi')
]


# Power Bundle
POWER: Bundle = [
    # Exponential Operator
    Operator(symbol='^', func=lambda a, b: a ** b, priority=Priority.IN_HIGH, kind='in', operands=2),
    Operator(symbol='**', func=lambda a, b: a ** b, priority=Priority.IN_HIGH, kind='in', operands=2),

    # Root Operator
    Operator(symbol='_/', func=lambda a: sqrt(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='_/', func=lambda a, b: b ** (1 / a), priority=Priority.IN_HIGH, kind='in', operands=2),

    # Logarithmic Operators
    Operator(symbol='log', func=lambda a, b: log(b, a), priority=Priority.PRE, kind='pre', operands=2),
    Operator(symbol='lg', func=lambda a: log10(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='ln', func=lambda a: log(a), priority=Priority.PRE, kind='pre', operands=1),

    # Tetration Operator
    Operator(symbol='^^', func=tetrate, priority=Priority.IN_HIGH, kind='in', operands=2)
]


# Trigonometry Bundle
TRIGONOMETRY: Bundle = [
    # Unit Conversion
    Operator(symbol='deg', func=lambda a: (pi / 180) * a, priority=Priority.POST, kind='post', operands=1),
    Operator(symbol='degrees', func=lambda a: (pi / 180) * a, priority=Priority.POST, kind='post', operands=1),

    # Basic Operators
    Operator(symbol='sin', func=lambda a: sin(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='sine', func=lambda a: sin(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='cos', func=lambda a: cos(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='cosine', func=lambda a: cos(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='tan', func=lambda a: tan(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='tangent', func=lambda a: tan(a), priority=Priority.PRE, kind='pre', operands=1),

    # Reciprocal Operators
    Operator(symbol='csc', func=lambda a: 1 / sin(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='sec', func=lambda a: 1 / cos(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='cot', func=lambda a: 1 / tan(a), priority=Priority.PRE, kind='pre', operands=1),

    # Inverse Operators
    Operator(symbol='asin', func=lambda a: asin(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='acos', func=lambda a: acos(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='atan', func=lambda a: atan(a), priority=Priority.PRE, kind='pre', operands=1),

    # Hyperbolic Operators
    Operator(symbol='sinh', func=lambda a: sinh(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='cosh', func=lambda a: cosh(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='tanh', func=lambda a: tanh(a), priority=Priority.PRE, kind='pre', operands=1),

    # Inverse Hyperbolic Operators
    Operator(symbol='asinh', func=lambda a: asinh(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='acosh', func=lambda a: acosh(a), priority=Priority.PRE, kind='pre', operands=1),
    Operator(symbol='atanh', func=lambda a: atanh(a), priority=Priority.PRE, kind='pre', operands=1)
]


# Factorial Bundle
FACTORIAL: Bundle = [
    # Regular Factorial Operator
    Operator(symbol='!', func=lambda a: factorial(a), priority=Priority.POST, kind='post', operands=1),

    # Subfactorial Operator
    Operator(symbol='!', func=lambda a: factorial(a) * sum(((-1) ** k) / factorial(k) for k in range(0, a + 1)), priority=Priority.PRE, kind='pre', operands=1),

    # Double Factorial Operator
    Operator(symbol='!!', func=double_factorial, priority=Priority.POST, kind='post', operands=1),

    # Superfactorial Operator
    Operator(symbol='$', func=superfactorial, priority=Priority.POST, kind='post', operands=1),

    # Hyperfactorial Operator
    Operator(symbol='#', func=hyperfactorial, priority=Priority.POST, kind='post', operands=1),
]


# Logic Bundle
LOGIC: Bundle = [
    # Operands
    Operand(value=0, symbol='false'),
    Operand(value=1, symbol='true'),

    # Boolean Operators
    Operator(symbol='~', func=lambda a: 1 if a == 0 else 0, priority=Priority.BOOLEAN_HIGH, kind='pre', operands=1),
    Operator(symbol='not', func=lambda a: 1 if a == 0 else 0, priority=Priority.BOOLEAN_HIGH, kind='pre', operands=1),
    Operator(symbol='|', func=lambda a, b: 1 if a == 1 or b == 1 else 0, priority=Priority.BOOLEAN_LOW, kind='in', operands=2),
    Operator(symbol='||', func=lambda a, b: 1 if a == 1 or b == 1 else 0, priority=Priority.BOOLEAN_LOW, kind='in', operands=2),
    Operator(symbol='or', func=lambda a, b: 1 if a == 1 or b == 1 else 0, priority=Priority.BOOLEAN_LOW, kind='in', operands=2),
    Operator(symbol='xor', func=lambda a, b: 1 if (a == 1 and b == 0) or (a == 0 and b == 1) else 0, priority=Priority.BOOLEAN_MID, kind='in', operands=2),
    Operator(symbol='&', func=lambda a, b: 1 if a == 1 and b == 1 else 0, priority=Priority.BOOLEAN_MID, kind='in', operands=2),
    Operator(symbol='&&', func=lambda a, b: 1 if a == 1 and b == 1 else 0, priority=Priority.BOOLEAN_MID, kind='in', operands=2),
    Operator(symbol='and', func=lambda a, b: 1 if a == 1 and b == 1 else 0, priority=Priority.BOOLEAN_MID, kind='in', operands=2),

    # Comparison Operators
    Operator(symbol='<', func=lambda a, b: int(a < b), priority=Priority.COMPARISON_HIGH, kind='in', operands=2),
    Operator(symbol='<=', func=lambda a, b: int(a <= b), priority=Priority.COMPARISON_HIGH, kind='in', operands=2),
    Operator(symbol='>', func=lambda a, b: int(a > b), priority=Priority.COMPARISON_HIGH, kind='in', operands=2),
    Operator(symbol='>=', func=lambda a, b: int(a >= b), priority=Priority.COMPARISON_HIGH, kind='in', operands=2),
    Operator(symbol='=', func=lambda a, b: int(a == b), priority=Priority.COMPARISON_LOW, kind='in', operands=2),
    Operator(symbol='==', func=lambda a, b: int(a == b), priority=Priority.COMPARISON_LOW, kind='in', operands=2),
    Operator(symbol='<>', func=lambda a, b: int(a != b), priority=Priority.COMPARISON_LOW, kind='in', operands=2),
    Operator(symbol='!=', func=lambda a, b: int(a != b), priority=Priority.COMPARISON_LOW, kind='in', operands=2),

    # Ternary/Conditional Operator
    Operator(symbol='if', func=lambda a, b, c: b if a == 1 else c, priority=Priority.PRE, kind='pre', operands=3),
    Operator(symbol='?', func=lambda a, b, c: b if a == 1 else c, priority=Priority.PRE, kind='pre', operands=3)
]

### FUTURE FEATURES ###

# Series Bundle
# SERIES: Bundle = [
#     Operator(symbol='sum', func=lambda a, b, c: a + b + c, priority=9)
# ]



from enum import IntEnum


class Priority(IntEnum):
    '''
    # (Enum) Priority

    ### Boolean Operators
    NOT: BOOLEAN_HIGH
    '''
    # Boolean Operators
    BOOLEAN_LOW = 1     # or, |, ||
    BOOLEAN_MID = 3     # and, &, &&
    BOOLEAN_HIGH = 5    # not, ~

    # Comparison Operators
    COMPARISON_LOW = 7  # =, ==, <>, !=
    COMPARISON_HIGH = 9 # <, <=, >=, >

    # In-fix Operators
    IN_LOW = 11         # +, -
    IN_MID = 13         # *, x, /, //, %
    IN_HIGH = 15        # **, ^, _/, ^^

    # Pre-fix Operators
    PRE = 17            # -, sin, lg,
                        # log

    # Implicit Operator
    IMPLICIT = 19       # .

    # Post-fix Operators
    POST = 21           # !, #, $, deg

    # Arbitrary Operator
    OPERAND = 30        # 1, 2.5, pi



from typing import Dict, List, Optional, Union, cast
from time import sleep

# Base Classes
from .node import Node
from .tree import Tree
from .operand_ import Operand
from .operator_ import Operator

# Key Components
from .tokenizer import Tokenizer
from .lexer import Lexer
from .parser import Parser

# Exceptions
from .exceptions import *

# Default Bundles
from .default_bundles import *

# Utilities
from .utils.io_utils import clear_console
from .utils.list_utils import find_element_where, index_first_element
from .utils.string_utils import concatenate_horizontally, simplify_expression
from string import ascii_lowercase
from termcolor import colored
from copy import copy, deepcopy


class ParseTree(Tree):
    def __init__(self,
                 bundles: List[Bundle] = [
                     BASICS,
                     INCREMENT,
                     MOD,
                     CONSTANTS,
                     POWER,
                     TRIGONOMETRY,
                     FACTORIAL,
                     LOGIC
                 ],
                 implicit_operator: Operator = Operator(
                     symbol='.', func=lambda a, b: a * b, priority=Priority.IMPLICIT, kind='in', operands=2),
                 **kwargs):
        super().__init__(**kwargs)
        self.expression = ''
        self.__last_build_expression = ''
        self.__token_lookup: Dict[str, Union[List[Operator], Operand]] = dict()
        self.__implicit_operator = implicit_operator
        self.__tokenizer = Tokenizer(token_lookup=self.__token_lookup)
        self.__lexer = Lexer(token_lookup=self.__token_lookup, implicit_operator=self.__implicit_operator)
        self.__parser = Parser(token_lookup=self.__token_lookup, implicit_operator=self.__implicit_operator)
        for bundle in bundles:
            self.register(bundle=bundle)

    def validate_registration(self, op: Union[Operand, Operator]):
        if not (isinstance(op, Operand) or isinstance(op, Operator)):
            raise RegistrationError(
                'Attempted to register invalid token. Please ensure that your class inherits either Operand or Operator accordingly.')
        elif isinstance(op, Operand):
            symbol = op.symbol
            prev_op = self.__token_lookup.get(symbol)
            if prev_op is not None:
                raise RegistrationWarning(
                    f'Symbol \'{symbol}\' already registered with the value {str(prev_op)}. The Operand for \'{symbol}\' will be replaced to have value {str(op)}.')
        elif isinstance(op, Operator):
            num_of_operands = op.operands
            kind = op.kind

            # Operator operands
            if type(num_of_operands) is not int or num_of_operands < 1:
                raise LexingError(
                    f'Operator for {str(op)} has invalid number of operands ({num_of_operands})')

            # Operator details
            if kind == 'post':
                if num_of_operands != 1:
                    raise LexingError(
                        'Post-fix only allowed for unary operators')
            elif kind == 'in':
                if num_of_operands != 2:
                    raise LexingError(
                        'In-fix only allowed for binary operators')
            elif kind == 'pre':
                pass
            else:
                raise RegistrationError(
                    f'Unknown kind \'{kind}\' encountered for operator \'{op.symbol}\'')

            # Operator overloading
            current_op = self.__token_lookup.get(op.symbol)
            if current_op is not None:
                if type(current_op) is list and len(current_op) >= 2:
                    raise RegistrationError(
                        f'Operator \'{op.symbol}\' aready fully overloaded')
                elif type(current_op) is list and len(current_op) == 1:
                    pass

    def register(self, bundle: Bundle) -> None:
        for op in bundle:
            self.validate_registration(op)
            if isinstance(op, Operand):
                self.__token_lookup[op.symbol] = op
            elif isinstance(op, Operator):
                if op.symbol not in self.__token_lookup:
                    self.__token_lookup[op.symbol] = [op]
                elif isinstance(self.__token_lookup[op.symbol], List):
                    cast(List[Operator],
                         self.__token_lookup[op.symbol]).append(op)
            else:
                raise RegistrationError(
                    f'Unknown Operand/Operator encountered: {type(op)}')

    def lex_token(self, token: Optional[str] = None) -> Union[Operand, Operator, List[Operator]]:
        if token is not None:
            if token in self.__token_lookup:
                if isinstance(self.__token_lookup[token], List):
                    return cast(List[Operator], deepcopy(self.__token_lookup[token]))
                else:
                    return cast(Operand, deepcopy(self.__token_lookup[token]))
            raise LexingError(f'Unknown token {token} encountered')
        else:
            return deepcopy(self.__implicit_operator)

    def deregister(self, symbol: Optional[str] = None):
        if symbol is None:
            self.__token_lookup = {}
        else:
            self.__token_lookup.pop(symbol)
        return self

    def read(self, expression: str, method: Literal['replace', 'append'] = 'replace') -> Tree:
        if method == 'replace':
            self.expression = expression
        else:
            self.expression += expression
        return self

    def tokenize(self, expression: str):
        return self.__tokenizer.tokenize(expression=expression, kind='eager')

    def lex(self, tokens: List[str]):
        return self.__lexer.lex(tokens=tokens, kind='eager')

    def parse(self, token_obj_list: List):
        return self.__parser.parse(token_obj_list=token_obj_list)

    def build_sub_tree(self, token_obj_list: List):
        for token_obj in token_obj_list:
            if type(token_obj) is list:
                t = ParseTree()
                sub_root, _ = t.build_sub_tree(token_obj_list=token_obj)
                if sub_root is None:
                    raise ParsingError('Empty parenthesis encountered')
                sub_root.priority += 20
                self.insert(sub_root)
            else:
                self.insert(token_obj)
        return self.root, self.currentPointer

    def lazy_build_sub_tree(self, token_obj_list: List):
        for token_obj in token_obj_list:
            if type(token_obj) is list:
                t = ParseTree()
                for tmp in t.lazy_build_sub_tree(token_obj_list=token_obj):
                    yield (self, *tmp)
                sub_root = t.root
                if sub_root is None:
                    raise ParsingError('Empty parenthesis encountered')
                sub_root.priority += 20
                self.insert(sub_root)
                yield (self,)
            else:
                self.insert(token_obj)
                yield (self,)

    def build(self):
        if self.expression.strip() != '':
            self.reset()
            tokens = self.tokenize(self.expression)
            lexed_tokens = self.lex(tokens)
            parsed_tokens = self.parse(lexed_tokens)
            self.build_sub_tree(parsed_tokens)
            self.__last_build_expression = self.expression
            return self
        else:
            raise BuildTreeError(
                'Attempted to build parse tree with empty expression')

    def __str__(self) -> str:
        if self.root is None or self.expression != self.__last_build_expression:
            self()
        return super().__str__()

    def str_without_building(self) -> str:
        return super().__str__()

    def validate_full_tree(self, node: Optional[Node]) -> None:
        if node is None:
            raise ReduceTreeError('Tree is empty')
        if not node.isFull() and isinstance(node, Operator):
            erroneous_operator = colored(node.symbol, color='red')
            actual = colored(str(len(node.children)), color='red')
            expected = node.operands
            raise ReduceTreeError(
                f'Parse tree is not full; {erroneous_operator} has {actual}/{expected} children')
        for c in node.children:
            self.validate_full_tree(node=c)

    def evaluate(self) -> Optional[Union[int, float]]:
        if self.root is None or self.expression != self.__last_build_expression:
            self()
        self.validate_full_tree(node=self.root)
        self.update(which='values')
        return self.root.value if self.root is not None else None

    def animate(self, seconds_per_frame: Union[int, float] = 2):
        return Scene(
            token_lookup=self.__token_lookup,
            implicit_operator=self.__implicit_operator,
            seconds_per_frame=seconds_per_frame
        ).animate(expression=self.expression)

    def __call__(self):
        return self.build()


class Scene:
    def __init__(self,
                 token_lookup: Dict,
                 implicit_operator: Operator,
                 seconds_per_frame: Union[int, float] = 1) -> None:
        self.token_lookup = token_lookup
        self.implicit_operator = implicit_operator
        self.seconds_per_frame = seconds_per_frame
        self.build_tree_ratio = 3
        self.reduce_tree_ratio = 3

    def stencil(self, expression, tokens=None, lexed=None, parsed=None, tree=None) -> None:
        clear_console()
        if tokens is None:
            process = 'TOKENIZATION'
        elif lexed is None:
            process = 'LEXING'
        elif parsed is None:
            process = 'PARSING'
        elif tree is None:
            process = 'BUILDING PARSE TREE'
        else:
            process = 'REDUCING PARSE TREE'
        print(f'CURRENT PROCESS: {process}')
        print()
        print(f'Expression: {expression}')
        if tokens is not None:
            print()
            print('TOKENS >>>')
            print(tokens)
        if lexed is not None:
            print()
            print('LEXED TOKENS >>>')
            print(lexed)
            print()
        if parsed is not None:
            print('PARSED TOKENS >>>')
            print(parsed)
            print()
        if tree is not None:
            print('COMPLETE PARSE TREE >>>')
            print(tree)
            print()

    def animate_tokenization(self, expression):
        tokens: List[str] = []
        for i, j, t in Tokenizer(token_lookup=self.token_lookup).tokenize(expression=expression, kind='lazy'):
            self.stencil(expression)
            print(' ' * (i + 12) + '\u2191' + ' ' * (j - i - 1) + '\u2191')
            print(' ' * (i + 12) + 'i' + ' ' * (j - i - 1) + 'j')
            print(t)
            tokens = t
            sleep(self.seconds_per_frame)
        return tokens

    def animate_lexing(self, expression, tokens):
        lexed = []
        prefix = 1
        for o, c, l, t in Lexer(token_lookup=self.token_lookup, implicit_operator=self.implicit_operator).lex(tokens=tokens, kind='lazy'):
            self.stencil(expression, tokens)
            arrow = '\u2191'
            print(' ' * prefix + f'{arrow:^{len(t) + 2}}')
            print(f'(: {o}; ): {c}')
            print(l)
            lexed = l
            sleep(self.seconds_per_frame)
            prefix += len(t) + 4
        return lexed

    def animate_parsing(self, lexed):
        return Parser(token_lookup=self.token_lookup, implicit_operator=self.implicit_operator).parse(lexed)

    def animate_building(self, expression, tokens, lexed, parsed):
        tree = ParseTree()
        prev_display = ''
        for trees in tree.lazy_build_sub_tree(parsed):
            next_display = concatenate_horizontally(map(lambda t: t.str_without_building(), filter(
                lambda t: t.root is not None, trees)), padding=2)
            if next_display == prev_display:
                continue
            self.stencil(expression, tokens, lexed, parsed)
            print(next_display)
            prev_display = next_display
            tree = trees[0]
            sleep(self.seconds_per_frame * self.build_tree_ratio)
        return tree

    def animate_reduction(self, expression, tokens, lexed, parsed, tree, tree_repr):
        reduced_tree = tree
        prev_display = ''
        for t in self.dfs_reduce(node=tree.root, tree=tree):
            if str(t) == prev_display:
                continue
            self.stencil(expression, tokens, lexed, parsed, tree_repr)
            prev_display = str(t)
            print(prev_display)
            sleep(self.seconds_per_frame * self.reduce_tree_ratio)
        self.stencil(expression, tokens, lexed, parsed, tree_repr)
        reduced_tree.root = Operand(value=reduced_tree.root())
        print(str(reduced_tree))
        return reduced_tree.root

    def dfs_reduce(self, node: Node, tree):
        for c in node.children:
            yield from self.dfs_reduce(c, tree)
        for i, c in enumerate(node.children):
            node.children[i] = Operand(value=c())
            node.children[i].parent = node
            tree.update(which='dimensions')
            yield tree

    def animate(self, expression):
        expression_stripped = simplify_expression(expression)
        tokens = self.animate_tokenization(expression_stripped)
        lexed = self.animate_lexing(expression_stripped, tokens)
        parsed = self.animate_parsing(lexed)
        t = self.animate_building(expression_stripped, tokens, lexed, parsed)
        complete_tree = str(t)
        result = self.animate_reduction(
            expression, tokens, lexed, parsed, t, complete_tree)
        self.stencil(expression_stripped, tokens, lexed, parsed, complete_tree)
        print('REDUCED PARSE TREE >>>')
        print(result)
        print()
        print('EVALUATION >>>')
        result = t.evaluate()
        print(expression, '=', round(result, 5))
        return result



from typing import Literal, Union
from termcolor import colored


# Expression Errors
class SyntaxError(Exception):
    def __str__(self) -> str:
        return f'{self.__class__.__name__}: {super().__str__()}'


class TokenizationError(SyntaxError):
    def __init__(self,
                 expression_simplified: str,
                 kind: Union[Literal['symbol'], Literal['token']],
                 i: int,
                 j: int,
                 *args: object) -> None:
        super().__init__('Unknown {} {} encountered in {}{}{}'.format(
            kind,
            expression_simplified[i:j],
            colored(text=expression_simplified[:i], color='green'),
            colored(text=expression_simplified[i:j], color='red'),
            expression_simplified[j:]
        ), *args)


class LexingError(SyntaxError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ParsingError(SyntaxError):
    pass


# Tree Errors
class TreeError(Exception):
    pass


class RegistrationWarning(Warning):
    pass


class RegistrationError(Exception):
    pass


class BuildTreeError(TreeError):
    pass


class UpdateTreeError(TreeError):
    pass


class ReduceTreeError(TreeError):
    pass


class TreeReprError(TreeError):
    pass



from copy import deepcopy
from .operand_ import Operand
from .operator_ import Operator
from .exceptions import LexingError, ParsingError
from .utils.list_utils import *
from typing import Dict, List, cast, Union


class Parser:
    def __init__(self,
                 token_lookup: Dict,
                 implicit_operator: Operator) -> None:
        self.token_lookup = token_lookup
        self.implicit_operator = implicit_operator

    def lex_token(self, token=None):
        if token is not None:
            if token in self.token_lookup:
                if isinstance(self.token_lookup[token], List):
                    return cast(List[Operator], deepcopy(self.token_lookup[token]))
                else:
                    return cast(Operand, deepcopy(self.token_lookup[token]))
            raise LexingError(f'Unknown token {token} encountered')
        else:
            return deepcopy(self.implicit_operator)

    def parse(self, token_obj_list: List, operand_expected: bool = True, index: int = 0, depth: int = 0):
        parsed_tokens: List = []
        while index < len(token_obj_list):
            token_obj = token_obj_list[index]
            if token_obj == ')':
                operand_expected = False
                if depth > 0:
                    return index, parsed_tokens, operand_expected
            elif token_obj == '(':
                if not operand_expected:
                    parsed_tokens.append(self.lex_token())
                terminating_index, nested_parsed_tokens, operand_expected = self.parse(
                    token_obj_list=token_obj_list, operand_expected=True, index=index + 1, depth=depth + 1)
                parsed_tokens.append(nested_parsed_tokens)
                index = terminating_index
            elif token_obj == '}':
                if depth > 0:
                    return index, parsed_tokens, operand_expected
            elif token_obj == '{':
                terminating_index, nested_parsed_tokens, operand_expected = self.parse(
                    token_obj_list=token_obj_list, operand_expected=True, index=index + 1, depth=depth + 1)
                parsed_tokens.append(nested_parsed_tokens)
                index = terminating_index
            elif type(token_obj) is list:
                # Operator overloaded
                token_obj = find_element_where(
                    ls=token_obj, condition=lambda el: el.kind == 'pre' if operand_expected else el.kind != 'pre')
                operand_expected = self.insert_token_obj(
                    parsed_tokens=parsed_tokens, token_obj=token_obj, operand_expected=operand_expected)
            elif isinstance(token_obj, Operator) and token_obj.kind == 'post':
                operand_expected = self.insert_token_obj(
                    parsed_tokens=parsed_tokens, token_obj=token_obj, operand_expected=operand_expected)
            else:
                operand_expected = self.insert_token_obj(
                    parsed_tokens=parsed_tokens, token_obj=token_obj, operand_expected=operand_expected)
            index += 1
        return parsed_tokens

    def insert_token_obj(self, parsed_tokens: List[Union[Operand, Operator]], token_obj: Union[Operand, Operator], operand_expected: bool) -> bool:
        if operand_expected and isinstance(token_obj, Operand):
            parsed_tokens.append(token_obj)
            return False
        elif operand_expected and isinstance(token_obj, Operator) and token_obj.kind == 'pre':
            parsed_tokens.append(token_obj)
            return True
        elif (not operand_expected) and isinstance(token_obj, Operator) and token_obj.kind == 'pre':
            # Implicit Multiplication
            parsed_tokens.append(cast(Operator, self.lex_token()))
            parsed_tokens.append(token_obj)
            return True
        elif (not operand_expected) and isinstance(token_obj, Operator) and token_obj.kind == 'post':
            parsed_tokens.append(token_obj)
            return operand_expected
        elif (not operand_expected) and isinstance(token_obj, Operator):
            parsed_tokens.append(token_obj)
            return True
        elif (not operand_expected) and isinstance(token_obj, Operand):
            # Implicit Multiplication
            parsed_tokens.append(cast(Operator, self.lex_token()))
            parsed_tokens.append(token_obj)
            return False
        elif operand_expected and isinstance(token_obj, Operator) and token_obj.kind != 'pre':
            raise ParsingError(
                f'Expected operand but received non pre-fix operator {token_obj}')
        else:
            raise ParsingError(
                'Unknown error encountered. Please check your expression again.')



from typing import Union
from time import sleep
import os


def clear_console() -> None:
    os.system('cls' if os.name == 'nt' else 'clear')


def variable_speed_print(message: str,
                         seconds_per_char: Union[int, float] = 0,
                         space_ratio: Union[int, float] = 0.1,
                         seconds_newline: Union[int, float] = 0) -> None:
    for i, char in enumerate(message):
        if char == '\n' and i < len(message) - 1 and message[i + 1] == '\n':
            sleep(seconds_newline)
        elif char == ' ':
            sleep(seconds_per_char * space_ratio)
        else:
            sleep(seconds_per_char)
        print(char, end='', flush=True)
    print()
